@@ -6,6 +6,7 @@ import os
 import time
 import warnings
 from datetime import timedelta
+from typing import Any
 
 import gradio as gr
 import httpx
@@ -15,10 +16,12 @@ from typing_extensions import assert_never
 from .. import utils
 from ..config import Config
 from .api import APIClient
+from .api import GPUSize
 from .api import AuthLevel
 from .api import QuotaInfos
 from .api import ScheduleResponse
-from .gradio import HTMLError
+from .gradio import info
+from .gradio import error
 from .gradio import get_event
 from .gradio import supports_auth
 
@@ -26,10 +29,10 @@ from .gradio import supports_auth
 TOKEN_HEADER = 'X-IP-Token'
 DEFAULT_SCHEDULE_DURATION = 60
 
-QUOTA_MESSAGE = "You have exceeded your GPU quota"
 UNUSED_MESSAGE = "GPU device not used"
 NO_GPU_MESSAGE_REGULAR = "No GPU was available"
 NO_GPU_MESSAGE_INQUEUE = "No GPU was available after 60s"
+EXAMPLES_RETRY_MESSAGE = "Try re-running outside of examples if it happened after clicking one"
 
 SIGNUP_ON_HF_TXT = "Create a free account"
 SIGNUP_ON_HF_URL = "https://huggingface.co/join"
@@ -43,10 +46,10 @@ def api_client():
     return APIClient(httpx_client)
 
 
-def startup_report():
+def startup_report(cgroup_path: str, gpu_size: GPUSize):
     retries, max_retries = 0, 2
     client = api_client()
-    while (status := client.startup_report()) is httpx.codes.NOT_FOUND: # pragma: no cover
+    while (status := client.startup_report(cgroup_path, gpu_size)) is httpx.codes.NOT_FOUND: # pragma: no cover
         time.sleep(1)
         if (retries := retries + 1) > max_retries:
             raise RuntimeError("Error while initializing ZeroGPU: NotFound")
@@ -85,8 +88,8 @@ def _toast_action(
         html += f'<a style="{style}" href="{link}">'
     html += text
     if supports_html:
-        html += '</a> '
-    html += desc
+        html += '</a>'
+    html += f" {desc}"
     markdown = f'[{text}]({link}) {desc}'
     return html, markdown
 
@@ -101,45 +104,60 @@ def schedule(
     if not (gradio_version := version.parse(gr.__version__)).major >= 4: # pragma: no cover
         raise RuntimeError("ZeroGPU is only compatible with Gradio 4+")
 
-    GRADIO_HTML_TOASTS = gradio_version.minor >= 39
+    GRADIO_HTML_TOASTS = gradio_version >= version.Version('4.39')
+    GRADIO_HANDSHAKE = gradio_version >= version.Version('5.16.1')
 
-    res, auth = api_client().schedule(
+    token, payload = _get_token_and_payload(request)
+    if token is not None and (token_error := payload.get('error')):
+        message = f"Falling back to IP-based quotas ({token_error})"
+        info("ZeroGPU client warning", message, level='warning')
+
+    res, meta = api_client().schedule(
         cgroup_path=utils.self_cgroup_device_path(),
         task_id=task_id,
-        token=_get_token(request),
+        token=token,
+        token_version=2 if GRADIO_HANDSHAKE else 1,
         duration_seconds=duration.seconds if duration is not None else None,
     )
 
+    auth = meta.auth
+
     if isinstance(res, ScheduleResponse):
         return res
 
     if isinstance(res, QuotaInfos): # pragma: no cover
         requested = duration.seconds if duration is not None else DEFAULT_SCHEDULE_DURATION
         if res.wait < timedelta(0):
-            raise gr.Error(
+            message = (
                 f"The requested GPU duration ({requested}s) "
                 f"is larger than the maximum allowed"
             )
-        else:
-            gpu = "Pro GPU" if auth == 'pro' else ("free GPU" if auth == 'regular' else "GPU")
+            raise error("ZeroGPU illegal duration", message)
+        elif token is None:
             message = (
-                f"You have exceeded your {gpu} quota "
-                f"({requested}s requested vs. {res.left}s left)."
+                f"Space app has reached its GPU limit. "
+                f"{EXAMPLES_RETRY_MESSAGE}"
             )
-            details_html, details_markdown = _toast_action(
-                auth=auth,
-                supports_html=GRADIO_HTML_TOASTS,
-                pro_message=f"Try again in {res.wait}",
-                unlogged_desc="to get more",
-                logged_desc="to get 5x more",
-                ending="usage quota",
-            )
-            message_html = f"{message} {details_html}"
-            message_text = f"{message} {details_markdown}"
-            raise HTMLError(html_string(message_html, message_text))
+            raise error("ZeroGPU quota exceeded", message)
+        else:
+            if payload.get('user') is None and res.wait == 0:
+                message = "You have exceeded your runs limit."
+            else:
+                gpu = "Pro GPU" if auth == 'pro' else ("free GPU" if auth == 'regular' else "GPU")
+                message = (
+                    f"You have exceeded your {gpu} quota "
+                    f"({requested}s requested vs. {res.left}s left). "
+                    f"Try again in {res.wait}"
+                )
+            raise error("ZeroGPU quota exceeded", message)
 
     if not isinstance(res, httpx.codes): # pragma: no cover
-        gr.Info("Waiting for a GPU to become available")
+        if meta.queuing_reason in ('node', None):
+            info("ZeroGPU queue", "Waiting for a GPU to become available")
+        elif meta.queuing_reason == 'concurrency':
+            info("ZeroGPU queue", "Waiting for a GPU slot on this Space")
+        else:
+            assert_never(meta.queuing_reason)
         # TODO: Sign-up message if not authenticated (after some time ?)
         connection_event = get_event()
         if connection_event is None and request is not None:
@@ -159,6 +177,9 @@ def schedule(
                     raise RuntimeError("Connection closed by visitor while queueing")
                 continue
             if event.event == 'failed':
+                if token is None:
+                    message = f"{NO_GPU_MESSAGE_INQUEUE}. {EXAMPLES_RETRY_MESSAGE}"
+                    raise error("ZeroGPU quota exceeded", message)
                 details_html, details_markdown = _toast_action(
                     auth=auth,
                     supports_html=GRADIO_HTML_TOASTS,
@@ -169,17 +190,21 @@ def schedule(
                 )
                 message_html = f"{NO_GPU_MESSAGE_INQUEUE}. {details_html}"
                 message_text = f"{NO_GPU_MESSAGE_INQUEUE} {details_markdown}"
-                raise HTMLError(html_string(message_html, message_text))
+                message = html_string(message_html, message_text)
+                raise error("ZeroGPU queue timeout", message, html=True)
             if event.event == 'succeeded':
                 assert event.data is not None
                 if connection_event is not None and not connection_event.alive:
                     release(event.data.allowToken)
                     raise RuntimeError("Connection closed by visitor on queue success")
-                gr.Info("Successfully acquired a GPU")
+                info("ZeroGPU queue", "Successfully acquired a GPU", level='success')
                 return event.data
 
     if res is httpx.codes.SERVICE_UNAVAILABLE:
-        raise gr.Error(NO_GPU_MESSAGE_REGULAR)
+        raise error("ZeroGPU client error", NO_GPU_MESSAGE_REGULAR)
+
+    if res is httpx.codes.UNAUTHORIZED: # pragma: no cover
+        raise error("ZeroGPU client error", "Expired ZeroGPU proxy token")
 
     # TODO: Find a way to log 'detail' response field
     raise RuntimeError(f"ZeroGPU API /schedule error: {res} ({httpx.codes.get_reason_phrase(res)})") # pragma: no cover
@@ -204,7 +229,7 @@ def release(
 
     if res is httpx.codes.NO_CONTENT: # pragma: no cover
         try:
-            gr.Warning(UNUSED_MESSAGE)
+            info("ZeroGPU client warning", UNUSED_MESSAGE, level='warning')
         except AttributeError:
             pass
         warnings.warn(UNUSED_MESSAGE, RuntimeWarning)
@@ -230,10 +255,21 @@ def _get_token(request: gr.Request | None) -> str | None:
 
     headers = getattr(request, 'headers', None)
     if headers is None or not hasattr(headers, '__dict__'):
-        raise gr.Error("Internal Gradio error")
+        raise error("ZeroGPU client error", "Internal Gradio error")
 
     # Compatibility trick
     if not hasattr(headers, 'get'):
         headers = headers.__dict__ # pragma: no cover
 
     return headers.get(TOKEN_HEADER.lower())
+
+
+def _get_token_and_payload(request: gr.Request | None) -> tuple[str | None, dict[str, Any]]:
+    if (token := _get_token(request)) is None:
+        return None, {}
+    try:
+        payload = utils.jwt_payload(token)
+    except Exception: # pragma: no cover
+        warnings.warn("Error while decoding X-IP-Token JWT")
+        return token, {}
+    return token, payload
