@@ -12,7 +12,6 @@ from concurrent.futures import ProcessPoolExecutor
 from concurrent.futures import ThreadPoolExecutor
 from contextlib import nullcontext
 from contextvars import copy_context
-from types import SimpleNamespace
 from typing import Any
 from typing import Callable
 
@@ -26,20 +25,14 @@ from torch.utils.weak import WeakTensorKeyDictionary
 from ...config import Config
 from ...utils import malloc_trim
 from ..tqdm import tqdm
-from . import bitsandbytes
+from . import cudart
 from .packing import ZeroGPUTensorPack
 from .packing import pack_tensors
 from .packing import pack_to_cuda
+from .static import *
 from .types import AliasId
 
 
-# Nvidia A100.80G MIG (drivers 535) / Torch 2.2.0
-CUDA_DEVICE_NAME = 'NVIDIA A100-SXM4-80GB MIG 3g.40gb'
-CUDA_TOTAL_MEMORY = 42144366592
-CUDA_MEM_GET_INFO = (41911451648, CUDA_TOTAL_MEMORY)
-CUDA_DEVICE_CAPABILITY = (8, 0)
-CUDA_DEVICE_PROPERTIES = SimpleNamespace(name=CUDA_DEVICE_NAME, major=8, minor=0, total_memory=CUDA_TOTAL_MEMORY, multi_processor_count=42)
-
 OPS_INPUTS_CHECK_NO_RETURN = (
     torch.Tensor.equal,
 )
@@ -53,15 +46,17 @@ OFFLOADED_ERROR_MESSAGE = "Cannot apply function {} on disk-offloaded Tensor {}"
 
 _tensor_make_subclass = torch.Tensor._make_subclass
 _asarray           = torch.asarray
+_device            = torch.device
 _cuda_init         = torch._C._cuda_init
 _cuda_exchange_device = torch.cuda._exchange_device
 _cuda_available      = torch.cuda.is_available
 _cuda_device_count   = torch.cuda.device_count
 _cuda_current_device = torch.cuda.current_device
-_cuda_mem_get_info   = torch.cuda.mem_get_info
 _cuda_get_device_capability   = torch.cuda.get_device_capability
 _cuda_get_device_properties   = torch.cuda.get_device_properties
 _cuda_get_device_name         = torch.cuda.get_device_name
+_cuda_memory_stats_as_nested_dict = torch.cuda.memory.memory_stats_as_nested_dict
+_cuda_cudart = torch.cuda.cudart
 
 # PyTorch 2.3
 _cuda_maybe_exchange_device = getattr(torch.cuda, '_maybe_exchange_device', None)
@@ -80,6 +75,15 @@ def empty_fake(tensor: torch.Tensor):
         fake = _tensor_make_subclass(tensor.__class__, fake, require_grad=tensor.requires_grad) # pyright: ignore [reportArgumentType]
     return fake
 
+# Torch 2.5: https://github.com/pytorch/pytorch/issues/144152
+def no_int_device(*args, **kwargs):
+    if len(args) and isinstance(index := args[0], int):
+        args = (f'cuda:{index}', *args[1:])
+    if isinstance(index := kwargs.get('device'), int):
+        kwargs['device'] = f'cuda:{index}'
+    return args, kwargs
+
+
 class ZeroGPUFunctionMode(TorchFunctionMode):
 
     def __torch_function__(self, func, types, args=(), kwargs: dict[str, Any] | None = None):
@@ -87,6 +91,7 @@ class ZeroGPUFunctionMode(TorchFunctionMode):
         kwargs = {} if kwargs is None else kwargs
 
         if func == torch._C._nn._parse_to:
+            args, kwargs = no_int_device(*args, **kwargs)
             return func(*args, **kwargs)
 
         # Redispatch: tensor.cuda() -> tensor.to(device='cuda')
@@ -99,7 +104,8 @@ class ZeroGPUFunctionMode(TorchFunctionMode):
 
         # Redispatch: tensor.to('cuda') -> tensor.to(device='cuda')
         if func == torch.Tensor.to and len(args) > 1:
-            device, dtype, _, memory_format = torch._C._nn._parse_to(*args[1:], **kwargs)
+            parse_to_args, parse_to_kwargs = no_int_device(*args[1:], **kwargs)
+            device, dtype, _, memory_format = torch._C._nn._parse_to(*parse_to_args, **parse_to_kwargs) # pyright: ignore [reportCallIssue, reportArgumentType]
             return self.__torch_function__(torch.Tensor.to, types, (args[0],), {
                 'device': device,
                 'dtype': dtype,
@@ -250,6 +256,15 @@ def _asarray_function_mode(*args, **kwargs):
     with torch._C.DisableTorchFunction():
         return function_mode.__torch_function__(_asarray, (), args=args, kwargs=kwargs)
 
+class _DeviceStringOnlyMeta(type):
+    def __instancecheck__(cls, instance):
+        return isinstance(instance, _device)
+
+class _DeviceStringOnly(metaclass=_DeviceStringOnlyMeta):
+    def __new__(cls, *args, **kwargs):
+        args, kwargs = no_int_device(*args, **kwargs)
+        return _device(*args, **kwargs)
+
 def _cuda_init_raise():
     raise RuntimeError(
         "CUDA must not be initialized in the main process "
@@ -271,19 +286,21 @@ def patch():
     torch.UntypedStorage.__new__ = _untyped_storage_new_register
     torch.UntypedStorage.device  = _untyped_storage_device # pyright: ignore [reportAttributeAccessIssue]
     torch.asarray           = _asarray_function_mode
+    torch.device            = _DeviceStringOnly
     torch._C._cuda_init     = _cuda_init_raise
     torch.cuda._exchange_device = _cuda_dummy_exchange_device
     torch.cuda.is_available   = lambda: True
     torch.cuda.device_count   = lambda: 1
     torch.cuda.current_device = lambda: 0
-    torch.cuda.mem_get_info   = lambda *args, **kwargs: CUDA_MEM_GET_INFO
     torch.cuda.get_device_capability = lambda *args, **kwargs: CUDA_DEVICE_CAPABILITY
     torch.cuda.get_device_properties = lambda *args, **kwargs: CUDA_DEVICE_PROPERTIES
     torch.cuda.get_device_name       = lambda *args, **kwargs: CUDA_DEVICE_NAME
+    torch.cuda.memory.memory_stats_as_nested_dict = lambda *args, **kwargs: CUDA_MEMORY_STATS_AS_NESTED_DICT
+    torch.cuda.cudart = lambda: cudart
     # PyTorch 2.3
     if _cuda_maybe_exchange_device is not None: # pragma: no cover
         setattr(torch.cuda, '_maybe_exchange_device', _cuda_dummy_exchange_device)
-    bitsandbytes.patch()
+    bitsandbytes().patch()
 
 def unpatch():
     try:
@@ -295,19 +312,21 @@ def unpatch():
     torch.UntypedStorage.__new__ = torch._C.StorageBase.__new__
     torch.UntypedStorage.device  = torch._C.StorageBase.device # pyright: ignore [reportAttributeAccessIssue]
     torch.asarray           = _asarray
+    torch.device            = _device
     torch._C._cuda_init     = _cuda_init
     torch.cuda._exchange_device = _cuda_exchange_device
     torch.cuda.is_available   = _cuda_available
     torch.cuda.device_count   = _cuda_device_count
     torch.cuda.current_device = _cuda_current_device
-    torch.cuda.mem_get_info   = _cuda_mem_get_info
     torch.cuda.get_device_capability = _cuda_get_device_capability
     torch.cuda.get_device_properties = _cuda_get_device_properties
     torch.cuda.get_device_name       = _cuda_get_device_name
+    torch.cuda.memory.memory_stats_as_nested_dict = _cuda_memory_stats_as_nested_dict
+    torch.cuda.cudart = _cuda_cudart
     # PyTorch 2.3
     if _cuda_maybe_exchange_device is not None: # pragma: no cover
         setattr(torch.cuda, '_maybe_exchange_device', _cuda_exchange_device)
-    bitsandbytes.unpatch()
+    bitsandbytes().unpatch()
 
 
 def _total_unpacked_size():
@@ -329,8 +348,9 @@ def _pack(offload_dir: str):
                 originals_dedup[original_id] = original
                 originals |= {original}
             fakes[originals_dedup[original_id]] += [fake]
+    total_size = _total_unpacked_size()
     progress = tqdm(
-        total=_total_unpacked_size(),
+        total=total_size,
         unit='B',
         unit_scale=True,
         desc="ZeroGPU tensors packing",
@@ -343,11 +363,13 @@ def _pack(offload_dir: str):
     for fake_list in fakes.values():
         for fake in fake_list:
             cuda_aliases[fake] = None
+    return total_size
 
 def pack():
-    _pack(Config.zerogpu_offload_dir)
+    total_size = _pack(Config.zerogpu_offload_dir)
     gc.collect()
     malloc_trim()
+    return total_size
 
 def init(nvidia_uuid: str):
     os.environ['CUDA_VISIBLE_DEVICES'] = nvidia_uuid
@@ -372,7 +394,7 @@ def _move(callback: Callable[[int]] | None = None):
     # Disk -> CUDA
     for tensor_pack in tensor_packs:
         pack_to_cuda(tensor_pack, callback=callback)
-    bitsandbytes.move()
+    bitsandbytes().move()
 
 def move(callback: Callable[[int]] | None = None):
     callback = callback if callback is not None else lambda _: None
@@ -384,3 +406,8 @@ def is_in_bad_fork():
     with ProcessPoolExecutor(mp_context=multiprocessing.get_context('fork')) as e:
         f = e.submit(torch.cuda._is_in_bad_fork)
         return f.result()
+
+def bitsandbytes():
+    # Lazy import
+    from . import bitsandbytes
+    return bitsandbytes
